import java.applet.Applet;
import java.awt.Container;
import java.awt.Graphics;
import java.awt.TextField;
import java.awt.Color;
import java.awt.Image;
import java.awt.Font;
import java.awt.Cursor;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseEvent;
import java.awt.event.KeyListener;
import java.awt.event.KeyEvent;

public class Mandelbrot extends Applet implements Runnable, MouseListener, MouseMotionListener, KeyListener
{

//___________________________________________________________________________ INITIALISATIONS

  Image image_buffer;

  int k;
  String k_text, k_input;

  Image image, tmp_image;
  int deb_x, deb_y, cur_x, cur_y;
  int fx, fy;
  int WIDTH, HEIGHT;
  double x, y, width, height;
  Color[] colors;
  boolean RUNNING, PAINTING, END;
  boolean HAS_HOME, HAS_PREV, HAS_NEXT;
  boolean SHOW_COMMANDS, DRAW_NEXT_SELECTION;
  boolean K_INPUT_ON;
  boolean RUN_AS_APPLICATION;
  Container _parent;
  double[] domains;
  int current_domain;
  int PROGRESS;

  public boolean RUNNING() { return RUNNING; }

  public void init()
  { addMouseListener(this);
    addMouseMotionListener(this);
    addKeyListener(this);
    k = 200;
    RUN_AS_APPLICATION = false;
    try
    { getCodeBase();
      if(getParameter("k") != null)
      { try { k = Integer.parseInt(getParameter("k")); }
        catch(Exception e) { k = 200; }
      }
    }
    catch(Exception e)
    { RUN_AS_APPLICATION = true;
      _parent = this;
      while(_parent.getParent() != null) _parent = _parent.getParent();
    }
    k_text = "" + k;
    k_input = "";
    init_colors();
    RUNNING = PAINTING = END = false;
    HAS_HOME = HAS_PREV = HAS_NEXT = false;
    SHOW_COMMANDS = DRAW_NEXT_SELECTION = K_INPUT_ON = false;
    WIDTH = size().width;
    HEIGHT = size().height - 20;
    image = createImage(WIDTH, HEIGHT);
    tmp_image = createImage(WIDTH, HEIGHT);
    deb_x = -1;
    cur_x = -1;
    fx = -1;
    init_domains();
    (new Thread(this)).start();
  }

  public void size_changed()
  { WIDTH = size().width;
    HEIGHT = size().height - 20;
    image = createImage(WIDTH, HEIGHT);
    tmp_image = createImage(WIDTH, HEIGHT);
    image_buffer = createImage(WIDTH, HEIGHT + 20);
    END = false;
    (new Thread(this)).start();
  }

  public void init_colors()
  { colors = new Color[k];
    colors[0] = new Color(0, 0, 0);
    int i = 1;
    int tour = 0;
    int c;
    while(i < k)
    { c = 5; while((i < k) && (i < (50  + (tour * 200)))) { colors[i] = new Color(0, c, 255 - c);       c += 5; i++; }
      c = 5; while((i < k) && (i < (100 + (tour * 200)))) { colors[i] = new Color(c, 255, 0);           c += 5; i++; }
      c = 5; while((i < k) && (i < (150 + (tour * 200)))) { colors[i] = new Color(255 - c, 255 - c, c); c += 5; i++; }
      c = 5; while((i < k) && (i < (200 + (tour * 200)))) { colors[i] = new Color(c, 0, 255);           c += 5; i++; }
      tour++;
    }
  }

  public void init_domains()
  { domains = new double[100];
    for(int i = 0; i < domains.length; i++) domains[i] = -1.0;
    current_domain = 0;
    set_home_domain();
    save_domain();
  }

  public void stop()  { END = true; }

//___________________________________________________________________________ DOMAINS

  public double[] home_domain()
  { double x      = -2.75;
    double width  = 4;
    double height = (double)HEIGHT / (double)WIDTH * width;
    double y      = -(height / 2);   
    return new double[]{x, y, width, height};
  }

  public void set_home_domain()
  { double[] home_domain = home_domain(); 
    x      = home_domain[0];
    y      = home_domain[1];
    width  = home_domain[2];
    height = home_domain[3];
  }

  public boolean is_home_domain()
  { double[] home_domain = home_domain();
    return x == home_domain[0] && y == home_domain[1] && width == home_domain[2] && height == home_domain[3];
  }

  public void save_domain()
  { domains[current_domain]     = x;
    domains[current_domain + 1] = y;
    domains[current_domain + 2] = width;
    domains[current_domain + 3] = height;
  }

  public void restor_domain()
  { x      = domains[current_domain];
    y      = domains[current_domain + 1];
    width  = domains[current_domain + 2];
    height = domains[current_domain + 3];
  }

  public void home()
  { new_domain();
    set_home_domain();
    save_domain();
    HAS_HOME = false;
    HAS_PREV = true;
    HAS_NEXT = false;
    reset_next_domains();
    (new Thread(this)).start();
  }

  public void prev()
  { current_domain -= 4;
    restor_domain();
    HAS_HOME = !is_home_domain();
    HAS_PREV = current_domain != 0;
    HAS_NEXT = true;
    (new Thread(this)).start();
  }

  public void next()
  { new_domain();
    restor_domain();
    HAS_HOME = !is_home_domain();
    HAS_PREV = true;
    HAS_NEXT = (current_domain + 4) < domains.length && domains[current_domain + 4] != -1.0;
    (new Thread(this)).start();
  }

  public void new_domain()
  { if(current_domain + 8 > domains.length)
    { double[] tmp = new double[domains.length + 100];
      for(int i = 0; i < domains.length; i++) tmp[i] = domains[i];
      for(int i = domains.length; i < tmp.length; i++) tmp[i] = -1.0;
      domains = tmp;
      tmp = null;
    }
    current_domain += 4;
  }

  public void reset_next_domains()
  { for(int i = current_domain + 4; i < domains.length && domains[i] != -1.0; i++) domains[i] = -1.0;
  }

//___________________________________________________________________________ MOUSE LISTENER

  public void mouseEntered(MouseEvent e) {}
  public void mouseExited(MouseEvent e)  {}

  public void mousePressed(MouseEvent e)
  { if(!RUNNING && e.getY() > 19 && !over_command(e))
    { deb_x = e.getX();
      deb_y = e.getY();
    }
    else deb_x = -1;
    if(RUN_AS_APPLICATION && e.getY() < 20 && !over_command(e))
    { fx = e.getX();
      fy = e.getY();
    }
    else fx = -1;
  }

  public void mouseReleased(MouseEvent e)
  { if(deb_x != -1 && cur_x > deb_x && cur_y > deb_y)
    { int selection[] = selection(deb_x, deb_y, cur_x, cur_y);
      double[] tr = translate(selection[0], selection[1] - 20, selection[2]);
      x      = tr[0];
      y      = tr[1];
      width  = tr[2];
      height = tr[3];
      new_domain();
      save_domain();
      HAS_HOME = true;
      HAS_PREV = true;
      HAS_NEXT = false;
      reset_next_domains();
      deb_x = -1;
      cur_x = -1;
      (new Thread(this)).start();
    }
    fx = fy = -1;
  }

  public void mouseClicked(MouseEvent e)
  { if(over_close(e)) System.exit(0);
    if(!RUNNING)
    { if(over_home(e)) home();
      else if(over_prev(e)) prev();
      else if(over_next(e)) next();
      else if(over_toggle_commands(e))
      { SHOW_COMMANDS = !SHOW_COMMANDS;
        K_INPUT_ON = !K_INPUT_ON;
        if(!K_INPUT_ON)
        { k_input = "";
          k_text = "" + k;
        }
        _repaint();
      }
      setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
    }
  }

  public void mouseMoved(MouseEvent e)
  { if((!RUNNING && over_command(e)) || (RUNNING && over_close(e))) setCursor(new Cursor(Cursor.HAND_CURSOR));
    else setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
    boolean OLD_DRAW_NEXT_SELECTION = DRAW_NEXT_SELECTION;
    DRAW_NEXT_SELECTION = SHOW_COMMANDS && !RUNNING && over_next(e);
    if(DRAW_NEXT_SELECTION != OLD_DRAW_NEXT_SELECTION) _repaint();
  }

  public void mouseDragged(MouseEvent e)
  { if(deb_x != -1)
    { cur_x = e.getX();
      cur_y = e.getY();
      repaint();
    }
    if(fx != -1)
    { _parent.setLocation
      ( (int)_parent.getLocation().x + e.getX() - fx,
        (int)_parent.getLocation().y + e.getY() - fy
      );
      fx = e.getX();
      fy = e.getY();
    }
  }

  public boolean over_command(MouseEvent e)
  { return over_home(e) || over_prev(e) || over_next(e) || over_toggle_commands(e) || over_close(e);
  }

  public boolean over_home(MouseEvent e)
  { return SHOW_COMMANDS && HAS_HOME && e.getX() > 12 && e.getX() < 22 && e.getY() > 31 && e.getY() < 41;
  }

  public boolean over_prev(MouseEvent e)
  { return SHOW_COMMANDS && HAS_PREV && e.getX() > 31 && e.getX() < 37 && e.getY() > 31 && e.getY() < 41;
  }

  public boolean over_next(MouseEvent e)
  { return SHOW_COMMANDS && HAS_NEXT && e.getX() > 49 && e.getX() < 55 && e.getY() > 31 && e.getY() < 41;
  }

  public boolean over_toggle_commands(MouseEvent e)
  { return e.getX() > 10 && e.getX() < 20 && e.getY() > 8 && e.getY() < 14;
  }

  public boolean over_close(MouseEvent e)
  { return RUN_AS_APPLICATION && e.getX() > WIDTH - 17 && e.getX() < WIDTH - 4 && e.getY() > 3 && e.getY() < 17;
  }

//___________________________________________________________________________ KEY LISTENER

  public void keyPressed(KeyEvent e)
  { if(K_INPUT_ON)
    { if
      ( e.getKeyChar() == '1' ||
        e.getKeyChar() == '2' ||
        e.getKeyChar() == '3' ||
        e.getKeyChar() == '4' ||
        e.getKeyChar() == '5' ||
        e.getKeyChar() == '6' ||
        e.getKeyChar() == '7' ||
        e.getKeyChar() == '8' ||
        e.getKeyChar() == '9' ||
        e.getKeyChar() == '0'
      ) update_k_input(e.getKeyChar());
      else if(e.getKeyCode() == KeyEvent.VK_BACK_SPACE) update_k_input('d');
      else if(e.getKeyCode() == KeyEvent.VK_ENTER)      update_k_input('a');
    }
  }

  public void keyReleased(KeyEvent e) {}
  public void keyTyped(KeyEvent e) {}

  public void update_k_input(char c)
  { if(c == 'd')
    { if(k_input.length() > 0) k_input = k_input.substring(0, k_input.length() - 1);
      k_text = k_input;
      repaint();
    }
    else
    { if(c == 'a')
      { k = Integer.parseInt(k_input);
        init_colors();
        k_input = "";
        (new Thread(this)).start();
      }
      else
      { if(k_input.length() == 0) k_text = "";
        k_input += c;
        k_text += c;
        repaint();
      }
    }
  }

//___________________________________________________________________________ COORDONNEES

  public double[] translate(int _x, int _y, int _width)
  { double[] tr = new double[4];
    /* x      */ tr[0] = x + (double)_x / WIDTH * width;
    /* width  */ tr[2] = (double)_width / WIDTH * width;
    /* height */ tr[3] = (double)HEIGHT / (double)WIDTH * tr[2];
    /* y      */ tr[1] = y + height - ((double)_y / HEIGHT * height) - tr[3];
    return tr;
  }

  public int[] next_selection()
  { int[] ns = new int[4];
    ns[0] = (int)((domains[current_domain + 4] - x) / width * WIDTH);
    ns[1] = (int)((y + height - domains[current_domain + 7] - domains[current_domain + 5]) / height * HEIGHT + 20);
    ns[2] = (int)(WIDTH * domains[current_domain + 6] / width);
    ns[3] = (int)(HEIGHT * ns[2] / WIDTH);  
    return ns;
  }

  public int[] selection(int x1, int y1, int x2, int y2)
  { int[] s = new int[4];
    s[0] = x1;
    s[1] = y1;
    if((x2 - x1) * 3 / 4 >= y2 - y1)
    { s[2] = x2 - x1;
      s[3] = (int)((double)HEIGHT * s[2] / WIDTH);
    }
    else
    { s[3] = y2 - y1;
      s[2] = (int)((double)WIDTH * s[3] / HEIGHT);
    }
    return s;
  }

//___________________________________________________________________________ PAINT

  public void repaint()
  { if(!PAINTING)
    { PAINTING = true;
      super.repaint();
    }
  }

  public void _repaint() {super.repaint(); }

  public void update(Graphics g)
  { if(image_buffer == null) image_buffer = createImage(size().width, size().height);
    paint(image_buffer.getGraphics());
    g.drawImage(image_buffer, 0, 0, this);
    PAINTING = false;
  }

  public void paint(Graphics g)
  { draw_domain(g);
    if(!RUNNING) draw_toggle_commands(g);
    if(RUN_AS_APPLICATION) draw_application_commands(g);
    if(RUNNING) draw_progress_bar(g);
    else if(image != null)
    { g.drawImage(image, 0, 20, this);
      if(deb_x != -1 && cur_x > deb_x && cur_y > deb_y) draw_selection(g);
      if(DRAW_NEXT_SELECTION) draw_next_selection(g);
      if(SHOW_COMMANDS) draw_navigation(g);
    }
  }

  public void draw_domain(Graphics g)
  { g.setColor(Color.black);
    g.fillRect(0, 0, WIDTH, 20);
    g.setColor(Color.white);
    g.setFont(new Font("Arial", Font.BOLD, 10));
    g.drawString("x:[" + x + "   " + (x + width) + "]      y:[" + y + "   " + (y + height) + "]", 30, 15);
  }

  public void draw_toggle_commands(Graphics g)
  { g.setColor(Color.white);
    for(int i = 0; i < 5; i++) g.drawLine(15 - i, 13 - i, 15 + i, 13 - i);
  }

  public void draw_application_commands(Graphics g)
  { g.setColor(new Color(175, 175, 175));
    g.fillRect(WIDTH - 16, 4, 12, 12);
    g.setColor(Color.black);
    g.drawLine(WIDTH - 13, 7, WIDTH - 7, 13);
    g.drawLine(WIDTH - 13, 13, WIDTH - 7, 7);
    g.setColor(new Color(100, 100, 100));
    g.drawLine(WIDTH - 16, 16, WIDTH - 4, 16);
    g.drawLine(WIDTH - 4, 4, WIDTH - 4, 16);
    g.setColor(Color.white);
    g.drawLine(WIDTH - 16, 4, WIDTH - 4, 4);
    g.drawLine(WIDTH - 16, 4, WIDTH - 16, 16);
  }

  public void draw_progress_bar(Graphics g)
  { g.drawImage(tmp_image, 0, 20, this);
    g.setColor(Color.white);
    g.fillRect(7, HEIGHT, WIDTH - 14, 15);
    g.setColor(new Color(28, 7, 143));
    g.drawRect(7, HEIGHT, WIDTH - 14, 15);
    g.fillRect(10, HEIGHT + 3, (int)((double)PROGRESS / WIDTH * (WIDTH - 20)), 10);
  }

  public void draw_selection(Graphics g)
  { g.setColor(Color.white);
    int selection[] = selection(deb_x, deb_y, cur_x, cur_y);
    g.drawRect(selection[0], selection[1], selection[2], selection[3]);
    double[] tr = translate(selection[0], selection[1] - 20, selection[2]);
    int info_y = selection[1] + selection[3] + 15;
    if(info_y > HEIGHT) info_y = selection[1] - 5;
    g.setFont(new Font("Arial", Font.BOLD, 10));
    g.drawString
    ( "x:[" + tr[0] + "   " + (tr[0] + tr[2]) + "]      y:[" + tr[1] + "   " + (tr[1] + tr[3]) + "]",
      10,
      info_y
    );
  }

  public void draw_next_selection(Graphics g)
  { g.setColor(Color.white);
    int[] next_selection = next_selection();
    g.drawRect(next_selection[0], next_selection[1], next_selection[2], next_selection[3]);
    int info_y = next_selection[1] + (int)((double)HEIGHT / (double)WIDTH * (next_selection[2])) + 15;
    if(info_y > HEIGHT) info_y = next_selection[1] - 5;
    g.setFont(new Font("Arial", Font.BOLD, 10));
    g.drawString
    ( "x:[" + domains[current_domain + 4]
      + "   " + (domains[current_domain + 4] + domains[current_domain + 6])
      + "]      y:[" + domains[current_domain + 5]
      + "   " + (domains[current_domain + 5] + domains[current_domain + 7]) + "]",
      10,
      info_y
    );
  }

  public void draw_navigation(Graphics g)
  { if(HAS_HOME || HAS_PREV || HAS_NEXT) 
    { g.setColor(Color.white);
      g.fillRect(10, 30, 50, 12);
      g.setColor(Color.black);
      g.drawRect(10, 30, 50, 12);
      if(HAS_HOME)
      { g.setColor(new Color(30, 90, 10));
        g.fillRect(13, 32, 9, 9);
      }
      if(HAS_PREV)
      { g.setColor(new Color(28, 7, 143));
        for(int i = 0; i < 5; i++) { g.drawLine(32 + i, 36 - i, 32 + i, 36 + i); }
      }
      if(HAS_NEXT)
      { g.setColor(new Color(28, 7, 143));
        for(int i = 0; i < 5; i++) { g.drawLine(54 - i, 36 - i, 54 - i, 36 + i); }
      }
    }
    g.setColor(Color.white);
    g.fillRect(10, 50, 50, 12);
    g.setColor(Color.black);
    g.drawRect(10, 50, 50, 12);
    g.setColor(Color.black);
    g.setFont(new Font("Arial", Font.BOLD, 10));
    g.drawString("" + k_text, 15, 60);
  }

//___________________________________________________________________________ CALCUL DES COULEURS

  public void run()
  { tmp_image.getGraphics().drawImage(image, 0, 0, this); 
    PROGRESS = 0;
    RUNNING = true;
    repaint();
    set_colors(image.getGraphics());
    RUNNING = false;
    _repaint();
  }

  public void set_colors(Graphics g)
  { double dx = width / WIDTH;
    double dy = height / HEIGHT;
    double c = x , ci;
    for(int i = 0; i < WIDTH && !END; i++)
    { ci = y;
      for(int j = HEIGHT - 1; j >= 0 && !END; j--)
      { g.setColor(colors[mandelbrot_rank(c, ci)]);
        g.drawLine(i, j, i, j);
        ci += dy;
      }
      PROGRESS++;
      c += dx;
      if(i % 10 == 0) repaint();
    }
  }

  public int mandelbrot_rank(double c, double ci)
  { int _k = 0;
    double z = 0, zi = 0, tmp_z;
    while((_k != k) && (Math.sqrt(z*z + zi*zi) < 2))
    { _k++;
      tmp_z = z*z - zi*zi + c;
      zi = 2*z*zi + ci;
      z = tmp_z;
    }
    if(_k == k) return 0;
    return _k;
  }

}
